//+--------------------------------------------------------------------+
//| File name:  MT4-FST Expert.mq4                                     |
//| Version:    0.20 Beta 2010-02-02                                   |
//| Copyright:   2009, Miroslav Popov - All rights reserved!          |
//| Website:    http://forexsb.com/                                    |
//| Support:    http://forexsb.com/forum/                              |
//| License:    Freeware under the following circumstances:            |
//|                                                                    |
//| This code is a part of Forex Strategy Trader. It is free for       |
//| use and distribution as an integral part of Forex Strategy Trader. |
//| One can modify it in order to improve the code or to fit it for    |
//| personal use. This code or any part of it cannot be used in        |
//| another applications without a permission. Contact information     |
//| cannot be changed.                                                 |
//|                                                                    |
//| NO LIABILITY FOR CONSEQUENTIAL DAMAGES                             |
//|                                                                    |
//| In no event shall the author be liable for any damages whatsoever  |
//| (including, without limitation, incidental, direct, indirect and   |
//| consequential damages, damages for loss of business profits,       |
//| business interruption, loss of business information, or other      |
//| pecuniary loss) arising out of the use or inability to use this    |
//| product, even if advised of the possibility of such damages.       |
//+--------------------------------------------------------------------+
#include <WinUser32.mqh>

#property copyright "Copyright  2009, Miroslav Popov"
#property link      "http://forexsb.com/"

#define EXPERT_VERSION           "0.20 Beta"
#define SERVER_SEMA_NAME         "MT4-FST Expert ID - "
#define TRADE_SEMA_NAME          "TradeIsBusy"
#define TRADE_SEMA_WAIT          100
#define TRADE_SEMA_TIMEOUT       10000
#define TRADE_RETRY_COUNT        4
#define TRADE_RETRY_WAIT         100

#define FST_REQ_ACCOUNT_INFO     3
#define FST_REQ_BARS             4
#define FST_REQ_ORDER_SEND       7
#define FST_REQ_ORDER_MODIFY     8
#define FST_REQ_ORDER_CLOSE      9
#define FST_REQ_PING             11
#define FST_REQ_MARKET_INFO_ALL  12
#define FST_REQ_TERMINAL_INFO    13

#define FST_ERR_INVALID_REQUEST  -1
#define FST_ERR_WRONG_ORD_TYPE   -10005
#define FST_ERR_POS_ALREADY_OPEN -10010
#define FST_ERR_NO_OPEN_POSITION -10015

#define OP_SQUARE                -1

#import "MT4-FST Library.dll"
string FST_LibraryVersion();
void FST_OpenConnection(int id);
void FST_CloseConnection(int id);
void FST_Ping(int id, string symbol, int period, int time, double bid, double ask, int spread, double rates[][6], int bars,
        double accountBalance, double accountEquity, double accountProfit, double accountFreeMargin, int positionTicket,
        int positionType, double positionLots, double positionOpenPrice, double positionStopLoss, 
        double positionTakeProfit, double positionProfit, string positionComment);
int FST_Tick(int id, string symbol, int period, int time, double bid, double ask, int spread, double rates[][6], int bars,
        double accountBalance, double accountEquity, double accountProfit, double accountFreeMargin, int positionTicket,
        int positionType, double positionLots, double positionOpenPrice, double positionStopLoss, 
        double positionTakeProfit, double positionProfit, string positionComment);
void FST_MarketInfoAll(int id, double point, double digits, double spread, double stoplevel, double lotsize,
        double tickvalue, double ticksize, double swaplong, double swapshort, double starting, double expiration,
        double tradeallowed, double minlot, double lotstep, double maxlot, double swaptype, double profitcalcmode,
        double margincalcmode, double margininit, double marginmaintenance, double marginhedged,
        double marginrequired, double freezelevel);
void FST_AccountInfo(int id, string name, int number, string company, string server, string currency, int leverage,
        double balance, double equity, double profit, double credit, double margin, double freemarginmode,
        double freemargin, int stopoutmode, int stopout);
void FST_TerminalInfo(int id, string symbol, string company, string path, string expertversion);
void FST_Bars(int id, string symbol, int period, double rates[][6], int bars);
int  FST_Request(int id, string &symbol[], int &iargs[], int icount, double &dargs[], int dcount);
void FST_Response(int id, int ok, int code);
#import

// Connection_ID serves to identify the expert when multiple copies of Forex Strategy Trader are used.
// It must be a unique number between 0 and 1000.
// The same number has to be entered in Forex Strategy Trader.
extern int  Connection_ID = 0;

// A unique number of the expeert's orders.
extern int Expert_Magic = 20011023;

// BreakEven_Activating_Distance determines at what profit the Stop Loss will be moved at BreakEven price.
// MarketInfo(symbol, MODE_STOPLEVEL) <= BreakEven_Activating_Distance <= 2000
// If BreakEven_Activating_Distance = 0, the BreakEven is not active.
// If BreakEven_Activating_Distance > 0, the Stop Loss will be placed at BreakEven once the price makes such a move. 
extern int BreakEven_Activating_Distance = 0;

// TrailingStop_Moving_Step determines the step of changing the Trailing Stop.
// 0 <= TrailingStop_Moving_Step <= 2000
// If TrailingStop_Moving_Step = 0, the Trailing Stop trails at every new extreme price in the position's direction.
// If TrailingStop_Moving_Step > 0, the Trailing Stop moves at steps equal to the number of pips chosen.
extern int TrailingStop_Moving_Step = 0;

bool     Separated_SL_TP  = false; // It's for brokers like FXCM and FXOpen that don't allow sending OrderSend(..) with SL and TP.
bool     IsServer         = false; // It shows whether the server is running.
bool     ConnectedToDLL   = false; // It shows whether the expert was connected to the dll.
double   TrailingStopPips = 0;     // It's used for the Trailing Stop indicator.
int      LastError        = 0;     // The number of last error.
bool     FST_Connected    = false; // Shows if FST is connected.
datetime TimeLastPing     = 0;     // Time of last ping from Forex Strategy Trader.

// Aggregate position.
int    PositionTicket     = 0;
int    PositionType       = OP_SQUARE;
double PositionLots       = 0;
double PositionOpenPrice  = 0;
double PositionStopLoss   = 0;
double PositionTakeProfit = 0;
double PositionProfit     = 0;
string PositionComment    = "";

///
/// Expert's initialization function.
///
int init()
{
    string message = "MT4-FST Expert version " + EXPERT_VERSION + " was started. An environment test is running...";
    Comment(message);
    Print(message);

    // Checks the requirements.
    if (!CheckEnvironment())
    {   // There is something wrong here. Exit.
        Sleep(20 * 1000);
        PostMessageA( WindowHandle( Symbol(), Period()), WM_COMMAND, 33050, 0);
        return (-1);
    }

    message = "The environment test was accomplished successfully.";
    Comment(message);
    Print(message);
    
    SetBrokersCompatibility();
    
    ReleaseTradeContext();
    
    Server(); // It's OK. We start the server.

    return (0);
}

///
/// Expert's start function.
/// 
int start()
{
    // We don't have to do anything here.
    // The new incoming ticks are processed by the server.
    return (0);
}

///
/// Expert's deinitialization function. 
///
int deinit()
{
   Comment("");
   if (ConnectedToDLL)
       FST_CloseConnection(Connection_ID);
   
   // Releases the global variable so the Expert
   // can be started on another chart.
   ReleaseServerSema();
   return (0);
}

///
/// Checks the working conditions.
///
bool CheckEnvironment()
{
    string message;
    
    // Checks if DLL is allowed.
    if (IsDllsAllowed() == false)
    {
        message = "\n" + "DLL call is not allowed." + "\n" +
                  "Please allow DLL loading in the MT options and restart the expert.";
        Comment(message);
        Print(message);
        return (false);
    }
    
    // Checks the Library version
    message = "\n" + "Cannot load \"MT4-FST Library.dll\"." + "\n" +
              "Please find more information about this error in the support forum.";
    Comment(message);
    string sLibraryVersion = FST_LibraryVersion();
    if (sLibraryVersion != "")
    {
        message = "MT4-FST Library version " + sLibraryVersion + " loaded successfully.";
        Comment(message);
        Print(message);
    }
    else
    {   // Meta Trader stops if it cannot load the dll.
        // Error 126 is displayed.
        return (false);
    }
    
    // Checks if you are logged in.
    if (AccountNumber() == 0)
    {
        message = "\n" + "You are not logged in. Please login first.";
        Comment(message);
        for (int attempt = 0; attempt < 200; attempt++)
        {
            if (AccountNumber() == 0)
                Sleep(300);
            else
                break;
        }
        if (AccountNumber() == 0)
            return (false);
    }
    
    // Checks the amount of bars available.
    int barsNecessary = MathMax(2 * 1440 / Period() + 10, 400); // Two days + 10 bars, min 400.
    if (!CheckChartBarsNumber(Symbol(), Period(), barsNecessary))
    {
        message = "\n" + "Cannot load enough bars! The expert needs minimum " + barsNecessary + " bars for this time frame." + "\n" + 
                  "Please load more data in the chart window and restart the expert.";
        Comment(message);
        Print(message);
        return (false);
    }
    
    // Checks the open positions.
    if (SetAggregatePosition(Symbol()) == -1)
    {   // Some arror with the current positions.
        return (false);
    }
    
    // Checks whether the expert was started on another chart.
    IsServer = GetServerSema();
    if (!IsServer)
    {
        string expert = "MT4-FST Expert ";
        if (Connection_ID > 0)
            expert = "MT4-FST Expert ID = " + Connection_ID + " ";

        message = "\n" + expert + "is already running on another chart!" + "\n" +
                  "Please stop all the instances and restart the expert." + "\n" + "\n" +
                  "If you cannot find the expert on the other charts, " + "\n" +  
                  "open the \"Global Variables\" window of MetaTrader (short key F3) and " + "\n" +
                  "delete the MT4-FST Expert ID - X variable. Restart the expert after that." + "\n" + "\n" +
                  "If you want to use the expert on several charts, you have to set a unique Connection_ID for " + "\n" + 
                  "all the experts and to do the same in FST. (The option must be switched on from Tools menu).";
        Comment(message);
        Print(message);
        return (false);
    }
    
    FST_OpenConnection(Connection_ID);
    ConnectedToDLL = true;
    
    // Everithing looks OK.
    return (true);
}

///
/// MT4-FST Server
///
int Server()
{
    string expertID = "";
    if (Connection_ID > 0)
        expertID = "(ID = " + Connection_ID + ") ";
    datetime tickTime = MarketInfo(Symbol(), MODE_TIME);
    string   message  = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Forex Strategy Trader is disconnected.";
    Comment(message);

    while (!IsStopped())
    {
        LastError = 0;
        RefreshRates();
        
        if (FST_Connected && (TimeLocal() - TimeLastPing > 60))
        {
            FST_Connected = false;
            message = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Forex Strategy Trader is disconnected.";
            Comment(message);
        }
        
        bool     bSucceed = false;
        string   symbol[1]; symbol[0] = Symbol();
        int      iargs[5];
        double   dargs[5];
        datetime time = MarketInfo(symbol[0], MODE_TIME);

        // Check for a new tick.
        if (time > tickTime)
        {
            SetAggregatePosition(symbol[0]);
            
            tickTime = time;

            if (BreakEven_Activating_Distance > 0)
                SetBreakEvenStop(symbol[0]);

            if (TrailingStopPips > 0)
                SetTrailingStop(symbol[0]);
        
            int iTickResponce = SendTick(symbol[0]);
            
            if (iTickResponce == 1)
            {
                FST_Connected = true;
                TimeLastPing  = TimeLocal();
                message = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Forex Strategy Trader is connected.";
            }
            else if (iTickResponce == 0)
            {
                FST_Connected = false;
                message = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Forex Strategy Trader is disconnected.";
            }
            else if (iTickResponce == -1)
            {
                message = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Error with sending a tick";
            }
            Comment(message);
            
            TimeLastPing = TimeLocal();
        }

        // Check for a request from Forex Strategy Trader.
        int request = FST_Request(Connection_ID, symbol, iargs, 5, dargs, 5);
        if (request < 0)
        {
            Comment("MT4-FST Library Server Error: ", request);
        }
        else if (request > 0)
        {
            //Print(symbol[0], " iargs[0]=", iargs[0], " iargs[1]=", iargs[1], " iargs[2]=", iargs[2], " iargs[3]=", iargs[3], " iargs[4]=", iargs[4],
            //                 " dargs[0]=", dargs[0], " dargs[1]=", dargs[1], " dargs[2]=", dargs[2], " dargs[3]=", dargs[3], " dargs[4]=", dargs[4]);
            switch (request)
            {
                case FST_REQ_PING:
                    // Forex Strategy Trader sends a ping.
                    GetPing(symbol[0]);
                    FST_Connected = true;
                    TimeLastPing  = TimeLocal();
                    
                    message = expertID + TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Forex Strategy Trader is connected.";
                    Comment(message);
                    break;
                    
                case FST_REQ_MARKET_INFO_ALL:
                    // Forex Strategy Trader requests full market info.
                    GetMarketInfoAll(symbol[0]);
                    break;
                    
                case FST_REQ_ACCOUNT_INFO:
                    // Forex Strategy Trader requests full account info.
                    FST_AccountInfo(Connection_ID, AccountName(), AccountNumber(), AccountCompany(), AccountServer(), AccountCurrency(),
                        AccountLeverage(), AccountBalance(), AccountEquity(), AccountProfit(), AccountCredit(), AccountMargin(),
                        AccountFreeMarginMode(), AccountFreeMargin(), AccountStopoutMode(), AccountStopoutLevel());
                    break;
                    
                case FST_REQ_TERMINAL_INFO:
                    // Forex Strategy Trader requests terminal info.
                    FST_TerminalInfo(Connection_ID, TerminalName(), TerminalCompany(), TerminalPath(), EXPERT_VERSION);
                    break;
                    
                case FST_REQ_BARS:
                    // Forex Strategy Trader requests historical bars.
                    GetBars(symbol[0], iargs[0], iargs[1]);
                    break;
                    
                case FST_REQ_ORDER_SEND:
                    // Forex Strategy Trader sends an order.
                    int orderResponse = ManageOrderSend(symbol[0], iargs[0], dargs[0], dargs[1], iargs[1], dargs[2], dargs[3], "", Expert_Magic, iargs[3]);
                    
                    if (orderResponse < 0)
                    {
                        int lastErrorOrdSend = GetLastError();
                        lastErrorOrdSend = IF_I(lastErrorOrdSend > 0, lastErrorOrdSend, LastError);
                        Print("Error in FST Request OrderSend: ", GetErrorDescription(lastErrorOrdSend));
                        FST_Response(Connection_ID, false, lastErrorOrdSend);
                    }
                    else 
                        FST_Response(Connection_ID, true, orderResponse);
                    break;

                case FST_REQ_ORDER_CLOSE:
                    // Forex Strategy Trader wants to close the current position.
                    bSucceed = CloseCurrentPosition(symbol[0], iargs[1]) == 0;
                    
                    int lastErrorOrdClose = GetLastError();
                    lastErrorOrdClose = IF_I(lastErrorOrdClose > 0, lastErrorOrdClose, LastError);
                    if (!bSucceed) Print("Error in OrderClose: ", GetErrorDescription(lastErrorOrdClose));
                    FST_Response(Connection_ID, bSucceed, lastErrorOrdClose);
                    break;

                case FST_REQ_ORDER_MODIFY:
                    // Forex Strategy Trader wants to modify the current position.
                    bSucceed = ModifyPosition(symbol[0], dargs[1], dargs[2]);
                    
                    int lastErrorOrdModify = GetLastError();
                    lastErrorOrdModify = IF_I(lastErrorOrdModify > 0, lastErrorOrdModify, LastError);
                    if (!bSucceed) Print("Error in OrderModify: ", GetErrorDescription(lastErrorOrdModify));
                    FST_Response(Connection_ID, bSucceed, lastErrorOrdModify);
                    break;

                default:
                    // Forex Strategy Trader doesn't know what to do.
                    FST_Response(Connection_ID, false, FST_ERR_INVALID_REQUEST);
                    
                    message = TimeToStr(TimeLocal(), TIME_DATE | TIME_SECONDS) + " Error - Forex Strategy Trader sent a wrong request.";
                    Comment(message);
                    Print(message);
                    break;
            }
        }

        Sleep(100);
    }
    
    return (0);
}

// ===========================================================================

///
/// Sets the virtual aggregate position.
/// Returns the number of open positions by FST or -1 if an error has occured.
int SetAggregatePosition(string symbol)
{
    PositionTicket     = 0;
    PositionType       = OP_SQUARE;
    PositionLots       = 0;
    PositionOpenPrice  = 0;
    PositionStopLoss   = 0;
    PositionTakeProfit = 0;
    PositionProfit     = 0;
    PositionComment    = "";

    int positions = 0;

    for (int i = 0; i < OrdersTotal(); i++) 
    {
        if (!OrderSelect(i, SELECT_BY_POS, MODE_TRADES))
        {
            Print("Error with OrderSelect: ", GetErrorDescription(GetLastError()));
            Comment("Cannot check current position!");
            continue;
        }

        if (OrderMagicNumber() != Expert_Magic || OrderSymbol() != symbol)
            continue; // An order not sent by Forex Strategy Trader.
     
        if (OrderType() == OP_BUYLIMIT || OrderType() == OP_SELLLIMIT || OrderType() == OP_BUYSTOP || OrderType() == OP_SELLSTOP)
            continue; // A pending order.

        if (PositionType >= 0 && PositionType != OrderType()) 
        {
            string message = "There are open positions in different directions!";
            Comment(message);
            Print(message);
            return (-1);
        }

        PositionTicket     = OrderTicket();
        PositionType       = OrderType();
        PositionOpenPrice  = (PositionLots * PositionOpenPrice + OrderLots() * OrderOpenPrice()) / (PositionLots + OrderLots());
        PositionLots      += OrderLots();
        PositionProfit    += OrderProfit();
        PositionStopLoss   = OrderStopLoss();
        PositionTakeProfit = OrderTakeProfit();
        PositionComment    = OrderComment();
        
        positions += 1;
    }
    
    if (PositionOpenPrice > 0)
        PositionOpenPrice = NormalizeDouble(PositionOpenPrice, MarketInfo(symbol, MODE_DIGITS));

    return (positions);
}

///
/// Sends correct order depending on the request and the current position.
///
int ManageOrderSend (string symbol, int type, double lots, double price, int slippage, double stoploss,
                     double takeprofit, string comment = "", int magic = 0, int expire = 0) 
{
    int orderResponse = -1;
    int positions = SetAggregatePosition(symbol);
    
    if (positions < 0)
        return (-1); // Error in SetAggregatePosition.

    if (positions == 0)
    {   // Open a new position.
        orderResponse = OpenNewPosition(symbol, type, lots, price, slippage, stoploss, takeprofit, magic);
    }
    else if (positions > 0)
    {   // There is a position.
        if ((PositionType == OP_BUY && type == OP_BUY) || (PositionType == OP_SELL && type == OP_SELL))
        {   // Add to the current position.
            orderResponse = AddToCurrentPosition(symbol, type, lots, price, slippage, stoploss, takeprofit, magic);
        }
        else if ((PositionType == OP_BUY && type == OP_SELL) || (PositionType == OP_SELL && type == OP_BUY))
        {
            if (MathAbs(PositionLots - lots) < MarketInfo(symbol, MODE_LOTSTEP) / 2)
            {   // The position's lots are equal to the opposite order's lots. We close the current position.
                orderResponse = CloseCurrentPosition(symbol, slippage);
            }
            else if (PositionLots > lots)
            {   // Reducing a position. (Partially closing).
                orderResponse = ReduceCurrentPosition(symbol, lots, price, slippage, stoploss, takeprofit, magic);
            }
            else if (PositionLots < lots)
            {   // Reversing a position.
                orderResponse = ReverseCurrentPosition(symbol, type, lots, price, slippage, stoploss, takeprofit, magic);
            }
        }
    }

    return (orderResponse);
}

///
/// Opens a new position at market price.
///
int OpenNewPosition (string symbol, int type, double lots, double price, int slippage, double stoploss, double takeprofit, int magic)
{
    int orderResponse = -1;
    
    if (type != OP_BUY && type != OP_SELL)
    {   // Error. Wrong order type!
        Print("Wrong 'Open new position' request - Wrong order type!");
        return (FST_ERR_WRONG_ORD_TYPE);
    }
    
    double orderLots = NormalizeEntrySize(symbol, lots);
    
    string comment = ""; 
    if (Connection_ID > 0) comment = "ID = " + Connection_ID;
    
    if (AccountFreeMarginCheck(symbol, type, orderLots) > 0)
    {
        if (Separated_SL_TP)
        {
            orderResponse = SendOrder(symbol, type, lots, price, slippage, 0, 0, comment, magic);
            if (orderResponse > 0)
            {
                double stopLossPrice   = GetStopLossPrice(symbol, type, orderLots, stoploss);
                double takeProfitPrice = GetTakeProfitPrice(symbol, type, takeprofit);
                orderResponse = ModifyPositionByTicket(symbol, orderResponse, stopLossPrice, takeProfitPrice);
            }
        }
        else
        {
            orderResponse = SendOrder(symbol, type, lots, price, slippage, stoploss, takeprofit, comment, magic);
            
            if (orderResponse < 0 && LastError == 130)
            {   // Invalid Stops. We'll check for forbiden direct set of SL and TP
                orderResponse = SendOrder(symbol, type, lots, price, slippage, 0, 0, comment, magic);
                if (orderResponse > 0)
                {
                    stopLossPrice   = GetStopLossPrice(symbol, type, orderLots, stoploss);
                    takeProfitPrice = GetTakeProfitPrice(symbol, type, takeprofit);
                    orderResponse   = ModifyPositionByTicket(symbol, orderResponse, stopLossPrice, takeProfitPrice);
                    if (orderResponse > 0)
                        Separated_SL_TP = true;
                }
            }
        }
    }
    
    SetAggregatePosition(symbol);
            
    return (orderResponse);
}

///
/// Adds more lots to the current position at the market price.
///
int AddToCurrentPosition(string symbol, int type, double lots, double price, int slippage, double stoploss, double takeprofit, int magic)
{
    // Checks if we have enough money. 
    if (AccountFreeMarginCheck(symbol, type, lots) <= 0)
        return (-1);

    int orderResponse = OpenNewPosition(symbol, type, lots, price, slippage, stoploss, takeprofit, magic);
    
    if(orderResponse < 0)
        return (orderResponse);

    double stopLossPrice   = GetStopLossPrice(symbol, type, PositionLots, stoploss);
    double takeProfitPrice = GetTakeProfitPrice(symbol, type, takeprofit);

    orderResponse = SetStopLossAndTakeProfit(symbol, stopLossPrice, takeProfitPrice);
    
    SetAggregatePosition(symbol);
    
    return (orderResponse);
}

///
/// Reduces current position at the market price.
///
int ReduceCurrentPosition(string symbol, double lots, double price, int slippage, double stoploss, double takeprofit, int magic)
{
    double newlots = PositionLots - lots;
    
    int orderstotal = OrdersTotal();
    int orders = 0;
    datetime openPos[][2];
    
    for (int i = 0; i < orderstotal; i++) 
    {
        if (!OrderSelect(i, SELECT_BY_POS, MODE_TRADES))
        {
            LastError = GetLastError();
            Print("Error in OrderSelect: ", GetErrorDescription(LastError));
            continue;
        }
        
        if (OrderMagicNumber() != magic || OrderSymbol() != symbol)
            continue;
        
        int orderType = OrderType();
        if (orderType != OP_BUY && orderType != OP_SELL)
            continue;
            
        orders++;
        ArrayResize(openPos, orders);
        openPos[orders - 1][0] = OrderOpenTime();
        openPos[orders - 1][1] = OrderTicket();
    }
    
    ArraySort(openPos, WHOLE_ARRAY, 0, MODE_ASCEND); 
        
    for (i = 0; i < orders; i++) 
    {
        OrderSelect(openPos[i][1], SELECT_BY_TICKET);
        double orderLots = IF_D(lots >= OrderLots(), OrderLots(), lots);
        CloseSelectedPosition(symbol, orderLots, slippage);
     
        lots -= orderLots;
        if (lots <= 0)
            break;
    }
   
    double stopLossPrice   = GetStopLossPrice(symbol, PositionType, newlots, stoploss);
    double takeProfitPrice = GetTakeProfitPrice(symbol, PositionType, takeprofit);
    
    int orderResponse = SetStopLossAndTakeProfit(symbol, stopLossPrice, takeProfitPrice);
    
    SetAggregatePosition(symbol);
    
    return (orderResponse);
}

///
/// Closes current position at market price.
/// Returns 0 if successful or negative if an error has occured.
///
int CloseCurrentPosition(string symbol, int slippage)
{
    int orderResponse = -1;
    int orderstotal   = OrdersTotal();
    int orders        = 0;
    datetime openPos[][2];
    
    for (int i = 0; i < orderstotal; i++) 
    {
        if (!OrderSelect(i, SELECT_BY_POS, MODE_TRADES))
        {
            LastError = GetLastError();
            Print("Error in OrderSelect: ", GetErrorDescription(LastError));
            continue;
        }
        
        if (OrderMagicNumber() != Expert_Magic || OrderSymbol() != symbol)
            continue;
        
        int orderType = OrderType();
        if (orderType != OP_BUY && orderType != OP_SELL)
            continue;
            
        orders++;
        ArrayResize(openPos, orders);
        openPos[orders - 1][0] = OrderOpenTime();
        openPos[orders - 1][1] = OrderTicket();
    }
    
    ArraySort(openPos, WHOLE_ARRAY, 0, MODE_ASCEND); 

    for (i = 0; i < orders; i++) 
    {
        OrderSelect(openPos[i][1], SELECT_BY_TICKET);
        orderResponse = CloseSelectedPosition(symbol, OrderLots(), slippage);
    }
    
    SetAggregatePosition(symbol);
    
    return (orderResponse);
}

///
/// Reverses current position at market price.
///
int ReverseCurrentPosition(string symbol, int type, double lots, double price, int slippage, double stoploss, double takeprofit, int magic)
{
    lots = lots - PositionLots;
    CloseCurrentPosition(symbol, slippage);
    
    int orderResponse = OpenNewPosition(symbol, type, lots, price, slippage, stoploss, takeprofit, magic);
    
    SetAggregatePosition(symbol);
    
    return (orderResponse);
}

///
/// Modify the position's Stop Loss and Take Profit
///
bool ModifyPosition(string symbol, double stoploss, double takeprofit)
{
    if (SetAggregatePosition(symbol) <= 0)
        return (false);
        
    double stopLossPrice   = GetStopLossPrice(symbol, PositionType, PositionLots, stoploss);
    double takeProfitPrice = GetTakeProfitPrice(symbol, PositionType, takeprofit);
    
    int responce = SetStopLossAndTakeProfit(symbol, stopLossPrice, takeProfitPrice);
    
    return (responce >= 0);
}

///
/// Sends an order to the broker.
///
int SendOrder(string symbol, int type, double lots, double price, int slippage, double stoploss, double takeprofit, string comment = "", int magic = 0)
{
    int orderResponse = -1;
    
    for (int attempt = 0; attempt < TRADE_RETRY_COUNT; attempt++)
    {
        if (!GetTradeContext())
            return (-1);
            
        double orderLots       = NormalizeEntrySize(symbol, lots);
        double orderPrice      = GetMarketPrice(symbol, type, price);
        double stopLossPrice   = GetStopLossPrice(symbol, type, orderLots, stoploss);
        double takeProfitPrice = GetTakeProfitPrice(symbol, type, takeprofit);
        color  colorDeal       = Lime; if (type == OP_SELL) colorDeal = Red;

        orderResponse = OrderSend(symbol, type, orderLots, orderPrice, slippage, stopLossPrice, takeProfitPrice, comment, magic, 0, colorDeal);
        
        ReleaseTradeContext();
        
        LastError = GetLastError();
        if (orderResponse > 0)
            break;
        
        if (LastError != 135 && LastError != 136 && LastError != 137 && LastError != 138)
            break;

        Print("Error with SendOrder: ", GetErrorDescription(LastError));

        Sleep(TRADE_RETRY_WAIT);
    }

    return (orderResponse);
}

///
/// Closes the currently selected position.
/// Returns 0 if successful or negative if an error has occured.
int CloseSelectedPosition(string symbol, double orderLots, int slippage)
{
    int orderType   = OrderType();
    int orderTicket = OrderTicket();

    for (int attempt = 0; attempt < TRADE_RETRY_COUNT; attempt++)
    {
        if (!GetTradeContext())
            return (-1);

        double orderPrice = IF_D(orderType == OP_BUY, MarketInfo(symbol, MODE_BID), MarketInfo(symbol, MODE_ASK));
        orderPrice = NormalizeDouble(orderPrice, Digits);
        bool responce = OrderClose(orderTicket, orderLots, orderPrice, slippage, Gold);
          
        ReleaseTradeContext();

        LastError = GetLastError();
        if (responce)
            return (0);
       
        Print("Error with ClosePositionByTicket: ", GetErrorDescription(LastError), ". Attempt No: ", (attempt + 1));
        Sleep(TRADE_RETRY_WAIT);
   }

    return (-1);
}

///
/// Updates the Stop Loss and the Take Profit of current positions. 
///
int SetStopLossAndTakeProfit(string symbol, double stopLossPrice, double takeProfitPrice)
{
    int responce = 1;
    
    for (int i = 0; i < OrdersTotal(); i++) 
    {
        if (!OrderSelect(i, SELECT_BY_POS, MODE_TRADES)) 
        {
            LastError = GetLastError();
            Print("Error with OrderSelect: ", GetErrorDescription(LastError));
            continue;
        }
        
        if (OrderMagicNumber() != Expert_Magic || OrderSymbol() != symbol)
            continue;
            
        int type = OrderType();
        if (type != OP_BUY && type != OP_SELL)
            continue;
            
        if (ModifyPositionByTicket(symbol, OrderTicket(), stopLossPrice, takeProfitPrice) < 0)
            responce = -1;
 
    }

    return (responce);
}

///
/// Modifies the position's Stop Loss and Take Profit.
///
int ModifyPositionByTicket(string symbol, int orderTicket, double stopLossPrice, double takeProfitPrice)
{
    if (!OrderSelect(orderTicket, SELECT_BY_TICKET)) 
    {
        LastError = GetLastError();
        Print("Error with OrderSelect: ", GetErrorDescription(LastError));
        return (-1);
    }
    
    stopLossPrice   = NormalizeDouble(stopLossPrice,   Digits);
    takeProfitPrice = NormalizeDouble(takeProfitPrice, Digits);

    double oldStopLoss   = NormalizeDouble(OrderStopLoss(),   Digits);
    double oldTakeProfit = NormalizeDouble(OrderTakeProfit(), Digits);

    if ((stopLossPrice == oldStopLoss) && (takeProfitPrice == oldTakeProfit))
        return (1); // There isn't anything to change.

    double orderOpenPrice = NormalizeDouble(OrderOpenPrice(), Digits);

    for (int attempt = 0; attempt < TRADE_RETRY_COUNT; attempt++)
    {
        if (!GetTradeContext())
            return (-1);
        bool rc = OrderModify(orderTicket, orderOpenPrice, stopLossPrice, takeProfitPrice, 0);
        ReleaseTradeContext();

        if (rc)
        {   // Modification was successful   
            return (1);
        }
        else
        {
            LastError = GetLastError();
            Print("Error with OrderModify(", orderTicket, ", ", orderOpenPrice, ", ", stopLossPrice, ", ", takeProfitPrice, ") ", GetErrorDescription(LastError), ".");
        }
    
        Sleep(TRADE_RETRY_WAIT);
    }

    return (-1);
}

///
/// Checks and corrects the order price.
///
double GetMarketPrice(string symbol, int type, double price)
{
    double orderPrice = price;
    
    if (type == OP_BUY)
        orderPrice = MarketInfo(symbol, MODE_ASK);
    else // if (type == OP_SELL)
        orderPrice = MarketInfo(symbol, MODE_BID);
        
    orderPrice = NormalizeDouble(orderPrice, MarketInfo(symbol, MODE_DIGITS));
    
    return (orderPrice);
}

///
/// Calculates the Take Profit price depending on the position direction and the Take Profit distance.
///
double GetTakeProfitPrice(string symbol, int type, double takeprofit)
{
    double takeProfitPrice = 0;

    if (takeprofit == 0)
        return (takeProfitPrice);

    if (takeprofit < MarketInfo(symbol, MODE_STOPLEVEL))
        takeprofit = MarketInfo(symbol, MODE_STOPLEVEL);
        
    if (type == OP_BUY)
        takeProfitPrice = MarketInfo(symbol, MODE_BID) + takeprofit * MarketInfo(symbol, MODE_POINT);
    else // if (type == OP_SELL)
        takeProfitPrice = MarketInfo(symbol, MODE_ASK) - takeprofit * MarketInfo(symbol, MODE_POINT);
    
    takeProfitPrice = NormalizeDouble(takeProfitPrice, MarketInfo(symbol, MODE_DIGITS));

    return (takeProfitPrice);
}

///
/// Calculates the Stop Loss price depending on the position direction and the Stop Loss distance.
///
double GetStopLossPrice(string symbol, int type, double lots, double stoploss)
{
    double stopLossPrice = 0;
    
    if (stoploss == 0)
        return (stopLossPrice);

    double correctedStopLoss = GetStopLossPips(symbol, stoploss, lots);
    if (correctedStopLoss < MarketInfo(symbol, MODE_STOPLEVEL))
        correctedStopLoss = MarketInfo(symbol, MODE_STOPLEVEL);

    if (type == OP_BUY)
        stopLossPrice = MarketInfo(symbol, MODE_BID) - correctedStopLoss * MarketInfo(symbol, MODE_POINT);
    else // if (type == OP_SELL)
        stopLossPrice = MarketInfo(symbol, MODE_ASK) + correctedStopLoss * MarketInfo(symbol, MODE_POINT);
    
    stopLossPrice = NormalizeDouble(stopLossPrice, MarketInfo(symbol, MODE_DIGITS));

    return (stopLossPrice);
}

/// Checks if we have regular Stop Loss or some kind of other stop.
/// The stop distance for a Trailing Stop is increased by 100000.
/// The stop distance for Account Percent Stop is increased by 200000.
double GetStopLossPips(string symbol, double stopLossPips, double lots)
{
    if (stopLossPips <= 100000)
    {   // Regular Stop Loss
        TrailingStopPips = 0;
    }
    else if (stopLossPips <= 200000)
    {   // Trailing Stop
        TrailingStopPips = stopLossPips - 100000;
        stopLossPips     = stopLossPips - 100000;
    }
    else if (stopLossPips > 200000)
    {   // Account Percent Stop
        TrailingStopPips = 0;
        double percent  = (stopLossPips - 200000) / 100;
        stopLossPips = AccountPercentStopPips(symbol, percent, lots);
    }

    return (stopLossPips);
}

///
/// The Account Percent Stop limits the loss to a 
/// predefined percent of the account balance. 
/// Returns the number of pips for the StopLoss.
double AccountPercentStopPips(string symbol, double percent, double lots)
{
    double balance  = AccountBalance();
    double exchrate = AccountExchangeRate(symbol);
    double lotsize  = MarketInfo(symbol, MODE_LOTSIZE);
    double point    = MarketInfo(symbol, MODE_POINT);
    double spread   = MarketInfo(symbol, MODE_SPREAD);

    double stopLossPips = percent * balance * exchrate / (lots * lotsize * point) - spread;

    return (stopLossPips);
}

/// Gets the Account Exchange Rate
/// AccountExchangeRate = AccountCurrency / QuotedCurrency
/// It serves to convert the profit of a deal in account currency.
/// This code has to be improved!
double AccountExchangeRate(string symbol)
{
    if (StringLen(symbol) != 6)
        return (1);

    string accCurrency    = AccountCurrency();
    string baseCurrency   = StringSubstr(symbol, 0, 3);
    string quotedCurrency = StringSubstr(symbol, 3, 3);

    if (accCurrency == quotedCurrency)
        return (1);
    else if (accCurrency == baseCurrency)
        return (MarketInfo(symbol, MODE_BID));
    else
    {
        string pair = StringConcatenate(accCurrency, quotedCurrency);
        double rate = MarketInfo(pair, MODE_BID);
        
        LastError = GetLastError();
        if (LastError == 4106)
        {
            pair = StringConcatenate(quotedCurrency, accCurrency);
            rate = MarketInfo(pair, MODE_BID);
            LastError = GetLastError();
            if (LastError == 0)
                rate = 1 / rate;
            else 
                rate = 1;
        }
        else if (LastError != 0)
            rate = 1;
        
        return (rate);
    }
        
    return (1);
}

///
/// Normalizes an entry order's size.
///
double NormalizeEntrySize(string symbol, double size)
{
    double minlot  = MarketInfo(symbol, MODE_MINLOT);
    double maxlot  = MarketInfo(symbol, MODE_MAXLOT);
    double lotstep = MarketInfo(symbol, MODE_LOTSTEP);

    if (size <= 0)
        return (0);
        
    double steps = NormalizeDouble((size - minlot) / lotstep, 0);
    size = minlot + steps * lotstep;

    if (size <= minlot)
        return (minlot);
        
    if (size >= maxlot)
        return (maxlot);
        
    return (size);
}

///
/// Sets a BreakEven stop of current positions. 
///
void SetBreakEvenStop(string symbol)
{
    if (SetAggregatePosition(symbol) <= 0)
        return;

    double breakeven = MarketInfo(symbol, MODE_STOPLEVEL);
    if (breakeven < BreakEven_Activating_Distance)
        breakeven = BreakEven_Activating_Distance;
    else if(BreakEven_Activating_Distance > 2000)
        breakeven = 2000;
    
    if (PositionType == OP_BUY)
    {
        double bid = MarketInfo(symbol, MODE_BID);
        if (bid - PositionOpenPrice >= Point * breakeven)
            if (PositionStopLoss < PositionOpenPrice)
                SetStopLossAndTakeProfit(symbol, PositionOpenPrice, PositionTakeProfit);
    }
    else if (PositionType == OP_SELL)
    {
        double ask = MarketInfo(symbol, MODE_ASK);
        if (PositionOpenPrice - ask >= Point * breakeven)
            if (PositionStopLoss == 0 || PositionStopLoss > PositionOpenPrice)
                SetStopLossAndTakeProfit(symbol, PositionOpenPrice, PositionTakeProfit);
    }
}

///
/// Sets a TrailingStop of current positions. 
///
void SetTrailingStop(string symbol)
{
    if (SetAggregatePosition(symbol) <= 0)
        return;
    
    // Validation.
    if (TrailingStop_Moving_Step > 2000)
        TrailingStop_Moving_Step = 2000;
        
    if (PositionType == OP_BUY)
    {
        double bid = MarketInfo(symbol, MODE_BID);
        if (bid - PositionOpenPrice >= Point * TrailingStopPips)
            if (PositionStopLoss < bid - Point * (TrailingStopPips + TrailingStop_Moving_Step))
                SetStopLossAndTakeProfit(symbol, bid - Point * TrailingStopPips, PositionTakeProfit);
    }   
    else if (PositionType == OP_SELL)
    {
        double ask = MarketInfo(symbol, MODE_ASK);
        if (PositionOpenPrice - ask >= Point * TrailingStopPips)
            if (PositionStopLoss > ask + Point * (TrailingStopPips + TrailingStop_Moving_Step))
                SetStopLossAndTakeProfit(symbol, ask + Point * TrailingStopPips, PositionTakeProfit);
    }
}

// ===========================================================================

///
/// Answers to a ping from Forex Strategy Trader.
///
void GetPing(string symbol)
{
    double rates[][6];
    if (ArrayCopyRates(rates) != Bars)
        return (false);

    SetAggregatePosition(symbol);
    
    double bid = MarketInfo(symbol, MODE_BID);
    double ask = MarketInfo(symbol, MODE_ASK);
    int spread = MathRound(MarketInfo(symbol, MODE_SPREAD));

    FST_Ping(Connection_ID, symbol, Period(), TimeCurrent(), bid, ask, spread, rates, Bars, 
        AccountBalance(), AccountEquity(), AccountProfit(), AccountFreeMargin(),
        PositionTicket, PositionType, PositionLots, PositionOpenPrice, PositionStopLoss, PositionTakeProfit, PositionProfit, PositionComment);
        
   return;
}

///
/// Sends a tick to Forex Strategy Trader.
///
int SendTick(string symbol)
{
    double rates[][6];
    if (ArrayCopyRates(rates) != Bars)
        return (false);

    SetAggregatePosition(symbol);
    
    double bid = MarketInfo(symbol, MODE_BID);
    double ask = MarketInfo(symbol, MODE_ASK);
    int spread = MathRound(MarketInfo(symbol, MODE_SPREAD));

    int responce = FST_Tick(Connection_ID, symbol, Period(), TimeCurrent(), bid, ask, spread, rates, Bars, 
        AccountBalance(), AccountEquity(), AccountProfit(), AccountFreeMargin(),
        PositionTicket, PositionType, PositionLots, PositionOpenPrice, PositionStopLoss, PositionTakeProfit, PositionProfit, PositionComment);
        
    return (responce);
}

///
/// Sends the bars to Forex Strategy Trader.
///
void GetBars(string symbol, int period, int barsNecessary)
{
    CheckChartBarsNumber(symbol, period, barsNecessary);
  
    RefreshRates();
    double rates[][6];
    int bars = ArrayCopyRates(rates, symbol, period);
    FST_Bars(Connection_ID, symbol, period, rates, bars);
}

//
// Checks if the chart contains enough bars.
//
bool CheckChartBarsNumber(string symbol, int period, int barsNecessary)
{
    int    bars = 0;
    double rates[][6];
    
    for (int attempt = 0; attempt < 10; attempt++)
    {
        RefreshRates();
        bars = ArrayCopyRates(rates, symbol, period);
        if (bars < barsNecessary && GetLastError() == 4066)
        {
            Comment("Loading...");
            Sleep(500);
        }
        else
            break;
            
        if (IsStopped())
            break;
    }
    
    if (bars < barsNecessary)
    {
        for (attempt = 0; attempt < 200; attempt++)
        {
            RefreshRates();
            bars = ArrayCopyRates(rates, symbol, period);
            if (bars < barsNecessary && GetLastError() != 4066)
            {
                Comment("There is not enough bars. Please press and hold Home key. ", bars, " bars loaded, ", barsNecessary, " bars necessary.");
                Sleep(300);
            }
            else
            {
                Comment("OK. Bars ", bars);
                break;
            }
            
            if (IsStopped())
                break;
        }
    }
    
    if (bars < barsNecessary)
    {
        string message = "There is not enough bars. FST needs minimum " + barsNecessary + " bars for this time frame. Currently " + bars + " bars are loaded.";
        Comment(message);
        Print(message);
    }

     return (bars >= barsNecessary);
}

///
/// Sends all the market information.
///
void GetMarketInfoAll(string symbol)
{
    FST_MarketInfoAll(Connection_ID,
                      MarketInfo(symbol, MODE_POINT             ),
                      MarketInfo(symbol, MODE_DIGITS            ),
                      MarketInfo(symbol, MODE_SPREAD            ),
                      MarketInfo(symbol, MODE_STOPLEVEL         ),
                      MarketInfo(symbol, MODE_LOTSIZE           ),
                      MarketInfo(symbol, MODE_TICKVALUE         ),
                      MarketInfo(symbol, MODE_TICKSIZE          ),
                      MarketInfo(symbol, MODE_SWAPLONG          ),
                      MarketInfo(symbol, MODE_SWAPSHORT         ),
                      MarketInfo(symbol, MODE_STARTING          ),
                      MarketInfo(symbol, MODE_EXPIRATION        ),
                      MarketInfo(symbol, MODE_TRADEALLOWED      ),
                      MarketInfo(symbol, MODE_MINLOT            ),
                      MarketInfo(symbol, MODE_LOTSTEP           ),
                      MarketInfo(symbol, MODE_MAXLOT            ),
                      MarketInfo(symbol, MODE_SWAPTYPE          ),
                      MarketInfo(symbol, MODE_PROFITCALCMODE    ),
                      MarketInfo(symbol, MODE_MARGINCALCMODE    ),
                      MarketInfo(symbol, MODE_MARGININIT        ),
                      MarketInfo(symbol, MODE_MARGINMAINTENANCE ),
                      MarketInfo(symbol, MODE_MARGINHEDGED      ),
                      MarketInfo(symbol, MODE_MARGINREQUIRED    ),
                      MarketInfo(symbol, MODE_FREEZELEVEL       ));

    return;
}

///
/// Customizes the expert for some brokers.
///
void SetBrokersCompatibility()
{
    // Chek broker.
    string broker = AccountCompany();
    
    // FXCM
    if (StringSubstr(broker, 0, 4) == "FXCM")
        Separated_SL_TP = true; 
    
    // FXOpen
    else if (StringSubstr(broker, 0, 6) == "FXOpen")
        Separated_SL_TP = true; 

    // ODL
    else if (StringSubstr(broker, 0, 3) == "ODL")
        Separated_SL_TP = true;
        
    // BenchMark
    else if (StringSubstr(broker, 0, 9) == "BenchMark")
        Separated_SL_TP = true;
        
    // Dukascopy
    else if (StringSubstr(broker, 0, 9) == "Dukascopy")
        Separated_SL_TP = true;
        
    return;
}

///
/// Checks the global variable in order to determine if
/// the server is running on another chart.
///
bool GetServerSema()
{
    if (GlobalVariableCheck(SERVER_SEMA_NAME + Connection_ID))
    {   // Global variable exists.
        double value = GlobalVariableGet(SERVER_SEMA_NAME + Connection_ID);
        if (value == 0)
        {   // Error in GlobalVariableGet.
            Print("Error in GlobalVariableGet: ", GetLastError());
            return (SaveServerSema());
        }
        else if (value < TimeLocal() - GetTickCount() / 1000)
        {   // Global variable has been set before Windows was started.
            return (SaveServerSema());
        }
        else
        {   // Server is working on another chart.
            return (false);
        }
    }

    // Global variable doesn't exist.
    return (SaveServerSema());
}

///
/// Releases the global variable when server stops.
///
void ReleaseServerSema()
{
    if (IsServer)
        GlobalVariableDel(SERVER_SEMA_NAME + Connection_ID);
}

///
/// Saves a global variable when server starts.
///
bool SaveServerSema()
{
    if (GlobalVariableSet(SERVER_SEMA_NAME + Connection_ID, TimeLocal()) != 0)
    {   // Global variable successfully set.
        return (true);
    }
    else
    {
        Print("Error in GlobalVariableSet: ", GetErrorDescription(GetLastError()));
        return (false);
    }
}

///
/// Sets a global variable to show that the trade content is busy. 
///
bool GetTradeContext()
{
    int start = GetTickCount();
    GetLastError();
    
    while (!GlobalVariableSetOnCondition(TRADE_SEMA_NAME, 1, 0)) 
    {
        int gle = GetLastError();
        if (gle != 0)
            Print("GTC: Error in GlobalVariableSetOnCondition: ", GetErrorDescription(gle));
        
        if (IsStopped())
            return (false); 
    
        if (GetTickCount() - start > TRADE_SEMA_TIMEOUT) 
        {
            Print("GTC: Timeout!");
            return (false);
        }
        
        Sleep(TRADE_SEMA_WAIT);
    }
    
    return (true);
}

///
/// Releases the global variable about the trade content.
///
void ReleaseTradeContext()
{
    while (!GlobalVariableSet(TRADE_SEMA_NAME, 0)) 
    {
        int gle = GetLastError();
        if (gle != 0)
            Print("RTC: Error in GlobalVariableSet: ", GetErrorDescription(gle));
        
        if (IsStopped())
            return; 
        
        Sleep(TRADE_SEMA_WAIT);
    }
}

///
/// Returns the text meaning of an error code. 
///
string GetErrorDescription(int iLastError)
{
    string sErrorDescription;

    switch(iLastError)
    {
        case 0:
        case 1:    sErrorDescription = "No error";                                                 break;
        case 2:    sErrorDescription = "Common error";                                             break;
        case 3:    sErrorDescription = "Invalid trade parameters";                                 break;
        case 4:    sErrorDescription = "Trade server is busy";                                     break;
        case 5:    sErrorDescription = "Old version of the client terminal";                       break;
        case 6:    sErrorDescription = "No connection with trade server";                          break;
        case 7:    sErrorDescription = "Not enough rights";                                        break;
        case 8:    sErrorDescription = "Too frequent requests";                                    break;
        case 9:    sErrorDescription = "Malfunctional trade operation (never returned error)";     break;
        case 64:   sErrorDescription = "Account disabled";                                         break;
        case 65:   sErrorDescription = "Invalid account";                                          break;
        case 128:  sErrorDescription = "Trade timeout";                                            break;
        case 129:  sErrorDescription = "Invalid price";                                            break;
        case 130:  sErrorDescription = "Invalid stops";                                            break;
        case 131:  sErrorDescription = "Invalid trade volume";                                     break;
        case 132:  sErrorDescription = "Market is closed";                                         break;
        case 133:  sErrorDescription = "Trade is disabled";                                        break;
        case 134:  sErrorDescription = "Not enough money";                                         break;
        case 135:  sErrorDescription = "Price changed";                                            break;
        case 136:  sErrorDescription = "Off quotes";                                               break;
        case 137:  sErrorDescription = "Broker is busy (never returned error)";                    break;
        case 138:  sErrorDescription = "Requote";                                                  break;
        case 139:  sErrorDescription = "Order is locked";                                          break;
        case 140:  sErrorDescription = "Long positions only allowed";                              break;
        case 141:  sErrorDescription = "Too many requests";                                        break;
        case 145:  sErrorDescription = "Modification denied because order too close to market";    break;
        case 146:  sErrorDescription = "Trade context is busy";                                    break;
        case 147:  sErrorDescription = "Expirations are denied by broker";                         break;
        case 148:  sErrorDescription = "Amount of open and pending orders has reached the limit";  break;
        case 149:  sErrorDescription = "Opening of an opposite position (hedging) is disabled";    break;
        case 150:  sErrorDescription = "An attempt to close a position contravening the FIFO rule";break;
        case 4000: sErrorDescription = "No error (never generated code)";                          break;
        case 4001: sErrorDescription = "Wrong function pointer";                                   break;
        case 4002: sErrorDescription = "Array index is out of range";                              break;
        case 4003: sErrorDescription = "No memory for function call stack";                        break;
        case 4004: sErrorDescription = "Recursive stack overflow";                                 break;
        case 4005: sErrorDescription = "Not enough stack for parameter";                           break;
        case 4006: sErrorDescription = "No memory for parameter string";                           break;
        case 4007: sErrorDescription = "No memory for temp string";                                break;
        case 4008: sErrorDescription = "Not initialized string";                                   break;
        case 4009: sErrorDescription = "Not initialized string in array";                          break;
        case 4010: sErrorDescription = "No memory for array\' string";                             break;
        case 4011: sErrorDescription = "Too long string";                                          break;
        case 4012: sErrorDescription = "Remainder from zero divide";                               break;
        case 4013: sErrorDescription = "Zero divide";                                              break;
        case 4014: sErrorDescription = "Unknown command";                                          break;
        case 4015: sErrorDescription = "Wrong jump (never generated error)";                       break;
        case 4016: sErrorDescription = "Not initialized array";                                    break;
        case 4017: sErrorDescription = "Dll calls are not allowed";                                break;
        case 4018: sErrorDescription = "Cannot load library";                                      break;
        case 4019: sErrorDescription = "Cannot call function";                                     break;
        case 4020: sErrorDescription = "Expert function calls are not allowed";                    break;
        case 4021: sErrorDescription = "Not enough memory for temp string returned from function"; break;
        case 4022: sErrorDescription = "System is busy (never generated error)";                   break;
        case 4050: sErrorDescription = "Invalid function parameters count";                        break;
        case 4051: sErrorDescription = "Invalid function parameter value";                         break;
        case 4052: sErrorDescription = "String function internal error";                           break;
        case 4053: sErrorDescription = "Some array error";                                         break;
        case 4054: sErrorDescription = "Incorrect series array using";                             break;
        case 4055: sErrorDescription = "Custom indicator error";                                   break;
        case 4056: sErrorDescription = "Arrays are incompatible";                                  break;
        case 4057: sErrorDescription = "Global variables processing error";                        break;
        case 4058: sErrorDescription = "Global variable not found";                                break;
        case 4059: sErrorDescription = "Function is not allowed in testing mode";                  break;
        case 4060: sErrorDescription = "Function is not confirmed";                                break;
        case 4061: sErrorDescription = "Send mail error";                                          break;
        case 4062: sErrorDescription = "String parameter expected";                                break;
        case 4063: sErrorDescription = "Integer parameter expected";                               break;
        case 4064: sErrorDescription = "Double parameter expected";                                break;
        case 4065: sErrorDescription = "Array as parameter expected";                              break;
        case 4066: sErrorDescription = "Requested history data in update state";                   break;
        case 4099: sErrorDescription = "End of file";                                              break;
        case 4100: sErrorDescription = "Some file error";                                          break;
        case 4101: sErrorDescription = "Wrong file name";                                          break;
        case 4102: sErrorDescription = "Too many opened files";                                    break;
        case 4103: sErrorDescription = "Cannot open file";                                         break;
        case 4104: sErrorDescription = "Incompatible access to a file";                            break;
        case 4105: sErrorDescription = "No order selected";                                        break;
        case 4106: sErrorDescription = "Unknown symbol";                                           break;
        case 4107: sErrorDescription = "Invalid price parameter for trade function";               break;
        case 4108: sErrorDescription = "Invalid ticket";                                           break;
        case 4109: sErrorDescription = "Trade is not allowed in the expert properties";            break;
        case 4110: sErrorDescription = "Longs are not allowed in the expert properties";           break;
        case 4111: sErrorDescription = "Shorts are not allowed in the expert properties";          break;
        case 4200: sErrorDescription = "Object is already exist";                                  break;
        case 4201: sErrorDescription = "Unknown object property";                                  break;
        case 4202: sErrorDescription = "Object is not exist";                                      break;
        case 4203: sErrorDescription = "Unknown object type";                                      break;
        case 4204: sErrorDescription = "No object name";                                           break;
        case 4205: sErrorDescription = "Object coordinates error";                                 break;
        case 4206: sErrorDescription = "No specified subwindow";                                   break;
        default:   sErrorDescription = "Unknown error";
    }

    return (sErrorDescription);
}

int IF_I(bool condition, int trueVal, int falseVal)
{
    if (condition)
        return (trueVal);
    else
        return (falseVal);
}

double IF_D(bool condition, double trueVal, double falseVal)
{
    if (condition)
        return (trueVal);
    else
        return (falseVal);
}